from .base_exception import BaseException


class DatasetException(BaseException):
    """Exception raised for errors in the dataset."""


class DatasetTaskError(DatasetException):
    """Exception raised for errors in the dataset task."""


class DatasetNotFoundError(DatasetException):
    """Exception raised for errors trying to load a dataset that does not exists."""


class DatasetAlreadyExistsError(DatasetException):
    """Exception raised for errors trying to create a dataset that already exists."""


class DatasetImportError(DatasetException):
    """Exception raised for errors trying to import a dataset."""


class DatasetAdapterError(DatasetException):
    """Exception raised for errors in the dataset adapter."""


class DatasetAdapterTaskError(DatasetAdapterError):
    """Exception raised for errors in the dataset adapter task."""


class DatasetEmptyError(DatasetAdapterError):
    """Exception raised for errors in the dataset adapter task."""


class DatasetSplitError(DatasetAdapterError):
    """Exception raised for errors in the dataset adapter task."""



from pathlib import Path
from typing import Optional, Union

from pydantic import BaseModel, Field, field_validator

from waffle_dough.field.base_field import BaseField
from waffle_dough.field.validator.image_validator import (
    validate_height,
    validate_split,
    validate_width,
)
from waffle_dough.type import SplitType, TaskType, get_split_types


class ImageInfo(BaseField):
    ext: str = Field(...)
    width: int = Field(...)
    height: int = Field(...)
    original_file_name: str = Field(...)
    date_captured: Optional[str] = Field(None)
    split: SplitType = Field(SplitType.UNSET.lower())

    def __init__(
        self,
        ext: str,
        width: int,
        height: int,
        original_file_name: str,
        task: Union[str, TaskType] = TaskType.AGNOSTIC,
        date_captured: Optional[str] = None,
        *args,
        **kwargs,
    ) -> "ImageInfo":
        """Image Information Field

        Args:
            task (Union[str, TaskType]): task type.
            ext (str): file ext(including .).
            width (int): width.
            height (int): height.
            original_file_name (str): original file name.
            task (Union[str, TaskType], optional): task type. Defaults to TaskType.AGNOSTIC.
            date_captured (Optional[str], optional): date captured. Defaults to None.

        Returns:
            ImageInfo: image information field.
        """
        super().__init__(
            task=task,
            ext=ext,
            width=width,
            height=height,
            original_file_name=original_file_name,
            date_captured=date_captured,
            *args,
            **kwargs,
        )

    @field_validator("width")
    def _check_width(cls, v):
        return validate_width(v)

    @field_validator("height")
    def _check_height_after(cls, v):
        return validate_height(v)

    @field_validator("split")
    def _check_split(cls, v):
        return validate_split(v)

    @classmethod
    def agnostic(
        cls,
        *,
        width: int,
        height: int,
        original_file_name: str,
        date_captured: Optional[str] = None,
        split: SplitType = SplitType.UNSET,
    ) -> "ImageInfo":
        """Agnostic Image Format

        Args:
            width (int): width.
            height (int): height.
            original_file_name (str): original file name
            date_captured (Optional[str], optional): date captured. Defaults to None.

        Returns:
            ImageInfo: image information field.
        """
        return cls(
            task=TaskType.AGNOSTIC,
            ext=Path(original_file_name).suffix.lower(),
            width=width,
            height=height,
            original_file_name=original_file_name,
            date_captured=date_captured,
            split=split,
        )


class UpdateImageInfo(BaseModel):
    split: Optional[SplitType] = Field(None)

    @field_validator("split")
    def _check_split(cls, v):
        return validate_split(v)

    @classmethod
    def agnostic(cls, *, split: SplitType = None) -> "UpdateImageInfo":
        """Agnostic Image Format

        Args:
            split (SplitType, optional): split. Defaults to None.

        Returns:
            UpdateImageInfo: update image information field.
        """
        return cls(
            split=split,
        )



from pathlib import Path
from typing import Union

from pycocotools.coco import COCO

from waffle_dough.exception import *
from waffle_dough.field import AnnotationInfo, CategoryInfo, ImageInfo
from waffle_dough.math.segmentation import convert_segmentation
from waffle_dough.type import SegmentationType, TaskType

from .base_adapter import BaseAdapter


class CocoAdapter(BaseAdapter):
    def __init__(
        self,
        images: dict[str, ImageInfo] = None,
        annotations: dict[str, AnnotationInfo] = None,
        categories: dict[str, CategoryInfo] = None,
        task: Union[str, TaskType] = TaskType.OBJECT_DETECTION,
        callbacks: list[callable] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, callbacks=callbacks, **kwargs)

        self.images: dict[str, ImageInfo] = images or {}
        self.annotations: dict[str, AnnotationInfo] = annotations or {}
        self.categories: dict[str, CategoryInfo] = categories or {}

        if task not in [
            TaskType.OBJECT_DETECTION,
            TaskType.INSTANCE_SEGMENTATION,
            TaskType.SEMANTIC_SEGMENTATION,
        ]:
            raise DatasetAdapterTaskError(f"Task {task} is not supported by COCO format.")
        self.task: Union[str, TaskType] = task.lower()

    @classmethod
    def from_target(
        cls,
        coco_dataset: Union[str, dict],
        task: Union[str, TaskType] = TaskType.OBJECT_DETECTION,
    ) -> "CocoAdapter":
        adapter = cls(task=task)

        if isinstance(coco_dataset, (str, Path)):
            coco = COCO(coco_dataset)
        elif isinstance(coco_dataset, dict):
            coco = COCO()
            coco.dataset = coco_dataset
            coco.createIndex()

        categories = {}
        coco_cat_id_to_new_cat_id = {}
        for cat_id, cat in coco.cats.items():
            cat_id = cat.pop("id", None) or cat_id
            cat = CategoryInfo.from_dict(
                task=task,
                d=cat,
            )
            coco_cat_id_to_new_cat_id[cat_id] = cat.id
            categories[cat.id] = cat

        images = {}
        coco_img_id_to_new_img_id = {}
        for img_id, img in coco.imgs.items():
            img = ImageInfo(
                ext=Path(img["file_name"]).suffix,
                width=img["width"],
                height=img["height"],
                original_file_name=img["file_name"],
                date_captured=img.get("date_captured", None),
                task=TaskType.AGNOSTIC,
            )
            coco_img_id_to_new_img_id[img_id] = img.id
            images[img.id] = img

        annotations = {}
        for ann in coco.anns.values():
            img = images[coco_img_id_to_new_img_id[ann["image_id"]]]
            cat = coco_cat_id_to_new_cat_id[ann["category_id"]]

            W = img.width
            H = img.height

            if task == TaskType.OBJECT_DETECTION:
                x1, y1, w, h = ann["bbox"]
                ann = AnnotationInfo.object_detection(
                    image_id=img.id,
                    category_id=cat,
                    bbox=[x1 / W, y1 / H, w / W, h / H],
                    iscrowd=getattr(ann, "iscrowd", None),
                    score=getattr(ann, "score", None),
                )
            elif task in [TaskType.INSTANCE_SEGMENTATION, TaskType.SEMANTIC_SEGMENTATION]:
                segmentation = ann["segmentation"]
                if isinstance(segmentation, dict):
                    segmentation = convert_segmentation(
                        segmentation, SegmentationType.RLE, SegmentationType.POLYGON
                    )

                for i, segmentation_ in enumerate(segmentation):
                    for j, point in enumerate(segmentation_):
                        segmentation[i][j] = point / W if j % 2 == 0 else point / H

                ann = AnnotationInfo.instance_segmentation(
                    image_id=img.id,
                    category_id=cat,
                    segmentation=segmentation,
                    bbox=getattr(ann, "bbox", None),
                    iscrowd=getattr(ann, "iscrowd", None),
                    score=getattr(ann, "score", None),
                )

            annotations[ann.id] = ann

        adapter.images = images
        adapter.annotations = annotations
        adapter.categories = categories

        return adapter

    def to_target(self, image_ids: list[str] = None, category_ids: list[str] = None) -> dict:
        coco = {}

        coco["images"] = []
        coco["categories"] = []
        coco["annotations"] = []

        target_img_id_to_coco_img_id = {}
        for img_id in image_ids or self.images.keys():
            img = self.images[img_id]
            coco_img_id = len(coco["images"]) + 1
            target_img_id_to_coco_img_id[img_id] = coco_img_id

            coco["images"].append(
                {
                    "id": coco_img_id,
                    "file_name": img.original_file_name,
                    "width": img.width,
                    "height": img.height,
                    "date_captured": img.date_captured,
                }
            )

        target_cat_id_to_coco_cat_id = {}
        for cat_id in category_ids or self.categories.keys():
            cat = self.categories[cat_id]
            coco_cat_id = len(coco["categories"]) + 1
            target_cat_id_to_coco_cat_id[cat_id] = coco_cat_id

            coco["categories"].append(
                {
                    "id": coco_cat_id,
                    "name": cat.name,
                    "supercategory": cat.supercategory,
                }
            )

        for ann_id, ann in self.annotations.items():
            img = self.images[ann.image_id]

            W = img.width
            H = img.height

            if self.task == TaskType.OBJECT_DETECTION:
                W = img.width
                H = img.height

                coco["annotations"].append(
                    {
                        "id": ann_id,
                        "image_id": target_img_id_to_coco_img_id[ann.image_id],
                        "category_id": target_cat_id_to_coco_cat_id[ann.category_id],
                        "bbox": [ann.bbox[0] * W, ann.bbox[1] * H, ann.bbox[2] * W, ann.bbox[3] * H],
                        "iscrowd": ann.iscrowd,
                        "score": ann.score,
                    }
                )
            elif self.task in [TaskType.INSTANCE_SEGMENTATION, TaskType.SEMANTIC_SEGMENTATION]:
                segmentation = ann.segmentation

                for i, segmentation_ in enumerate(segmentation):
                    for j, point in enumerate(segmentation_):
                        segmentation[i][j] = point * W if j % 2 == 0 else point * H

                coco["annotations"].append(
                    {
                        "id": ann_id,
                        "image_id": target_img_id_to_coco_img_id[ann.image_id],
                        "category_id": target_cat_id_to_coco_cat_id[ann.category_id],
                        "bbox": [ann.bbox[0] * W, ann.bbox[1] * H, ann.bbox[2] * W, ann.bbox[3] * H],
                        "segmentation": segmentation,
                        "iscrowd": ann.iscrowd,
                        "score": ann.score,
                    }
                )

        return coco



import logging
import uuid
from pathlib import Path
from typing import Union

import numpy as np
from PIL import Image, ImageDraw, ImageFont
from waffle_utils.file.network import get_file_from_url

from waffle_dough.exception import BaseException
from waffle_dough.field import AnnotationInfo, CategoryInfo, ImageInfo
from waffle_dough.image import io
from waffle_dough.type import TaskType

FONT_URL = "https://raw.githubusercontent.com/snuailab/assets/main/waffle/fonts/gulim.ttc"
FONT_NAME = "gulim.ttc"
BASE_FONT_SIZE = 15


# random colors with 1000 categories
colors = [
    [255, 30, 30],
    [30, 255, 30],
    [30, 30, 255],
    [255, 255, 30],
    [255, 30, 255],
    [30, 255, 255],
    [200, 100, 20],
    [20, 200, 100],
    [100, 20, 200],
    [160, 0, 220],
    [220, 160, 0],
    [0, 220, 160],
] + np.random.randint(0, 255, (1000, 3), dtype="uint8").tolist()


def _get_font(scale: float = 1.0) -> ImageFont.ImageFont:
    if not Path(FONT_NAME).exists():
        get_file_from_url(FONT_URL, FONT_NAME, True)
    return ImageFont.truetype(FONT_NAME, int(BASE_FONT_SIZE * scale))


def _convert_to_pil(image: np.ndarray) -> Image.Image:
    if image.dtype == np.uint8:
        return Image.fromarray(image)
    elif image.dtype == np.float32:
        return Image.fromarray((image * 255).astype(np.uint8))
    else:
        raise BaseException("image dtype must be uint8 or float32")


def _convert_to_numpy(image: Image.Image) -> np.ndarray:
    return np.array(image)


def visualize_classification(
    image: np.ndarray,
    annotations: list[AnnotationInfo],
    category_dict: dict[CategoryInfo],
    image_info: ImageInfo,
    color_map: dict[str, list[int]],
    scale: float = 1.0,
    opacity: float = 1.0,
) -> np.ndarray:

    pil_image = _convert_to_pil(image)
    draw = ImageDraw.Draw(pil_image, "RGBA")
    font = _get_font(scale)

    for i, annotation in enumerate(annotations):
        category = category_dict[annotation.category_id]
        color = tuple(color_map[annotation.category_id])
        text = f"{category.name}" + (f": {annotation.score}" if annotation.score is not None else "")

        text_box = draw.textbbox((0, 0 + 20 * i), text, font=font)
        draw.rectangle(
            text_box, fill=color + (int(255 * opacity),), outline=color, width=round(2 * scale)
        )
        draw.text(text_box[:2], text, (255, 255, 255), font=font)

    return _convert_to_numpy(pil_image)


def visualize_regression(
    image: np.ndarray,
    annotations: list[AnnotationInfo],
    category_dict: dict[CategoryInfo],
    image_info: ImageInfo,
    color_map: dict[str, list[int]],
    scale: float = 1.0,
    opacity: float = 1.0,
) -> np.ndarray:

    pil_image = _convert_to_pil(image)
    draw = ImageDraw.Draw(pil_image, "RGBA")
    font = _get_font(scale)

    for i, annotation in enumerate(annotations):
        category = category_dict[annotation.category_id]
        color = tuple(color_map[annotation.category_id])
        text = f"{category.name}: {annotation.value}"

        text_box = draw.textbbox((0, 0 + 20 * i), text, font=font)
        draw.rectangle(
            text_box, fill=color + (int(255 * opacity),), outline=color, width=round(2 * scale)
        )
        draw.text(text_box[:2], text, (255, 255, 255), font=font)

    return _convert_to_numpy(pil_image)


def visualize_object_detection(
    image: np.ndarray,
    annotations: list[AnnotationInfo],
    category_dict: dict[CategoryInfo],
    image_info: ImageInfo,
    color_map: dict[str, list[int]],
    scale: float = 1.0,
    opacity: float = 0.2,
) -> np.ndarray:

    pil_image = _convert_to_pil(image)
    draw = ImageDraw.Draw(pil_image, "RGBA")
    font = _get_font(scale)

    W, H = image_info.width, image_info.height

    for i, annotation in enumerate(annotations):
        category = category_dict[annotation.category_id]
        color = tuple(color_map[annotation.category_id])
        text = f"{category.name}" + (f": {annotation.score}" if annotation.score is not None else "")

        x1, y1, w, h = annotation.bbox
        x2, y2 = x1 + w, y1 + h
        x1, y1, x2, y2 = round(x1 * W), round(y1 * H), round(x2 * W), round(y2 * H)

        draw.rectangle(
            (x1, y1, x2, y2),
            fill=color + (int(127 * opacity),),
            outline=color,
            width=round(2 * scale),
        )

        text_box = draw.textbbox((x1, y1), text, font=font)
        th = text_box[3] - text_box[1]
        th = round(th * 1.25)
        pad = round(th * 0.1)
        text_box = (text_box[0], text_box[1] - th - pad, text_box[2], text_box[3] - th + pad)
        draw.rectangle(text_box, fill=color, outline=color, width=round(2 * scale))
        draw.text(text_box[:2], text, (255, 255, 255), font=font)

    return _convert_to_numpy(pil_image)


def visualize(
    image: np.ndarray,
    annotations: Union[AnnotationInfo, list[AnnotationInfo], dict, list[dict]],
    category_dict: dict[str, CategoryInfo],
    image_info: Union[ImageInfo, dict] = None,
    task: Union[TaskType, str] = None,
    color_map: dict[str, list[int]] = None,
    opacity: float = None,
) -> np.ndarray:
    """
    Visualize image with annotations and categories

    Args:
        image (np.ndarray): image to visualize
        image_infos (ImageInfo): image info to visualize
        annotations (Union[AnnotationInfo, list[AnnotationInfo]]): annotations to visualize
        category_dict (dict[str, CategoryInfo]): category dict to visualize
        task (Union[TaskType, str], optional): task type. Defaults to None.
        color_map (dict[str, list[int]], optional): color map that has category id as key and color as value. Defaults to None.
        opacity (float, optional): opacity of visualized image. Defaults to None.

    Returns:
        np.ndarray: visualized image
    """

    result = image.copy()

    # parse arguments
    ## get image info
    if image_info is not None:
        image_info = ImageInfo.from_dict(image_info) if isinstance(image_info, dict) else image_info
    else:
        image_info = ImageInfo(
            ext=".jpg",
            width=image.shape[1],
            height=image.shape[0],
            original_file_name=f"{uuid.uuid4()}.jpg",
        )

    ## get annotations
    annotations = annotations if isinstance(annotations, list) else [annotations]
    annotations = [
        AnnotationInfo.from_dict(annotation) if isinstance(annotation, dict) else annotation
        for annotation in annotations
    ]

    if len(annotations) == 0:
        return result

    ## get task type
    if task is None:
        task = TaskType.from_str(annotations[0].task)
    elif isinstance(task, str):
        task = TaskType.from_str(task)

    # visualize
    ## get color map
    if color_map is None:
        color_map = {}
        for i, category_id in enumerate(category_dict):
            color_map[category_id] = colors[i]

    ## get adaptive font scale
    scale = min(image.shape[0], image.shape[1]) / 256

    if task == TaskType.CLASSIFICATION:
        result = visualize_classification(
            result,
            annotations,
            category_dict,
            image_info,
            color_map,
            scale,
            **{"opacity": opacity} if opacity is not None else {},
        )
    elif task == TaskType.OBJECT_DETECTION:
        result = visualize_object_detection(
            result,
            annotations,
            category_dict,
            image_info,
            color_map,
            scale,
            **{"opacity": opacity} if opacity is not None else {},
        )
    elif task == TaskType.REGRESSION:
        result = visualize_regression(
            result,
            annotations,
            category_dict,
            image_info,
            color_map,
            scale,
            **{"opacity": opacity} if opacity is not None else {},
        )

    return result



from .base_exception import BaseException


class FieldException(BaseException):
    """Base class for exceptions in `Field` module."""


class FieldValidationError(FieldException):
    """Exception raised for errors in the field validation."""


class FieldTaskError(FieldException):
    """Exception raised for errors in the task validation."""


class FieldMissingError(FieldException):
    """Exception raised for errors in the missing fields."""



import logging
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

from waffle_utils.file import io

from waffle_dough.exception.database_exception import *
from waffle_dough.field import (
    AnnotationInfo,
    CategoryInfo,
    ImageInfo,
    UpdateAnnotationInfo,
    UpdateCategoryInfo,
    UpdateImageInfo,
)
from waffle_dough.type.split_type import SplitType

from .engine import create_session
from .repository import (
    annotation_repository,
    category_repository,
    image_repository,
)

logger = logging.getLogger(__name__)


class DatabaseService:
    def __init__(self, db_url: str, image_directory: Union[str, Path]):
        self.Session = create_session(db_url)
        self.image_directory = Path(image_directory)

    # Create
    def add_image(self, image: str, image_info: ImageInfo) -> ImageInfo:
        with self.Session() as session:
            io.copy_file(image, self.get_image_path(image_info), create_directory=True)
            try:
                image_info = image_repository.create(session, image_info)
            except Exception as e:
                io.remove_file(self.get_image_path(image_info))
                raise e

        return ImageInfo.model_validate(image_info, from_attributes=True)

    def add_category(self, category_info: CategoryInfo) -> CategoryInfo:
        with self.Session() as session:
            category = category_repository.create(session, category_info)

        return CategoryInfo.model_validate(category, from_attributes=True)

    def add_annotation(self, annotation_info: AnnotationInfo) -> AnnotationInfo:
        with self.Session() as session:
            annotation = annotation_repository.create(session, annotation_info)

        return AnnotationInfo.model_validate(annotation, from_attributes=True)

    # Read (Unit)
    def get_image(self, image_id: str) -> ImageInfo:
        with self.Session() as session:
            image = image_repository.get(session, image_id)

        if image is None:
            raise DatabaseNotFoundError(f"image does not exist: {image_id}")

        return ImageInfo.model_validate(image, from_attributes=True)

    def get_image_path(self, image_info: ImageInfo) -> Path:
        return Path(self.image_directory, f"{image_info.id}{image_info.ext}")

    def get_category(self, category_id: str) -> CategoryInfo:
        with self.Session() as session:
            category = category_repository.get(session, category_id)

        if category is None:
            raise DatabaseNotFoundError(f"category does not exist: {category_id}")

        return CategoryInfo.model_validate(category, from_attributes=True)

    def get_annotation(self, annotation_id: str) -> AnnotationInfo:
        with self.Session() as session:
            annotation = annotation_repository.get(session, annotation_id)

        if annotation is None:
            raise DatabaseNotFoundError(f"annotation does not exist: {annotation_id}")

        return AnnotationInfo.model_validate(annotation, from_attributes=True)

    # Read (Multi)
    def _get_query_dict(
        self,
        id: Union[str, list[str]] = None,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        annotation_id: Union[str, list[str]] = None,
        split: SplitType = None,
        filter_by: dict[str, Any] = {},
        filter_in: dict[str, Optional[list]] = {},
        filter_like: list[Tuple[str, str]] = [],
        order_by: List[Tuple[str, str]] = [],
        skip: int = None,
        limit: int = None,
    ) -> dict[str, Any]:
        query_dict = {}

        filter_by = filter_by.copy()
        filter_in = filter_in.copy()
        filter_like = filter_like.copy()
        order_by = order_by.copy()

        if id:
            filter_in.update({"id": [id] if isinstance(id, str) else id})

        if image_id:
            filter_in.update({"image_id": [image_id] if isinstance(image_id, str) else image_id})

        if category_id:
            filter_in.update(
                {"category_id": [category_id] if isinstance(category_id, str) else category_id}
            )

        if annotation_id:
            filter_in.update(
                {"id": [annotation_id] if isinstance(annotation_id, str) else annotation_id}
            )

        if split:
            filter_by.update({"split": split})

        query_dict.update(
            {
                "filter_by": filter_by,
                "filter_in": filter_in,
                "filter_like": filter_like,
                "order_by": order_by,
                "skip": skip,
                "limit": limit,
            }
        )

        return query_dict

    def _get_multi(self, repository, **kwargs) -> List[Any]:
        with self.Session() as session:
            objs = repository.get_multi(session, **self._get_query_dict(**kwargs))

        return objs

    def _get_count(self, repository, **kwargs) -> int:
        with self.Session() as session:
            count = repository.get_count(session, **self._get_query_dict(**kwargs))

        return count

    def get_images(
        self, image_id: Union[str, list[str]] = None, split: SplitType = None, **kwargs
    ) -> dict[str, ImageInfo]:
        images = self._get_multi(image_repository, id=image_id, split=split, **kwargs)
        return {image.id: ImageInfo.model_validate(image, from_attributes=True) for image in images}

    def get_categories(
        self, category_id: Union[str, list[str]] = None, **kwargs
    ) -> dict[str, CategoryInfo]:
        categories = self._get_multi(category_repository, id=category_id, **kwargs)
        return {
            category.id: CategoryInfo.model_validate(category, from_attributes=True)
            for category in categories
        }

    def get_annotations(
        self, annotation_id: Union[str, list[str]] = None, **kwargs
    ) -> dict[str, AnnotationInfo]:
        annotations = self._get_multi(annotation_repository, id=annotation_id, **kwargs)
        return {
            annotation.id: AnnotationInfo.model_validate(annotation, from_attributes=True)
            for annotation in annotations
        }

    # Read with relations
    def get_images_by_category_id(
        self, category_id: Union[str, list[str]], split: SplitType = None
    ) -> dict[str, ImageInfo]:
        annotations = self._get_multi(annotation_repository, category_id=category_id)
        image_ids = list(set(map(lambda annotation: annotation.image_id, annotations)))
        images = self._get_multi(image_repository, id=image_ids, split=split)
        return {image.id: ImageInfo.model_validate(image, from_attributes=True) for image in images}

    def get_images_by_annotation_id(
        self, annotation_id: Union[str, list[str]], split: SplitType = None
    ) -> dict[str, ImageInfo]:
        annotations = self._get_multi(annotation_repository, annotation_id=annotation_id)
        image_ids = list(set(map(lambda annotation: annotation.image_id, annotations)))
        images = self._get_multi(image_repository, id=image_ids, split=split)
        return {image.id: ImageInfo.model_validate(image, from_attributes=True) for image in images}

    def get_categories_by_annotation_id(
        self, annotation_id: Union[str, list[str]]
    ) -> dict[str, CategoryInfo]:
        annotations = self._get_multi(annotation_repository, annotation_id=annotation_id)
        category_ids = list(set(map(lambda annotation: annotation.category_id, annotations)))
        categories = self._get_multi(category_repository, id=category_ids)
        return {
            category.id: CategoryInfo.model_validate(category, from_attributes=True)
            for category in categories
        }

    def get_categories_by_image_id(self, image_id: Union[str, list[str]]) -> dict[str, CategoryInfo]:
        annotations = self._get_multi(annotation_repository, image_id=image_id)
        category_ids = list(set(map(lambda annotation: annotation.category_id, annotations)))
        categories = self._get_multi(category_repository, id=category_ids)
        return {
            category.id: CategoryInfo.model_validate(category, from_attributes=True)
            for category in categories
        }

    def get_annotations_by_image_id(
        self, image_id: Union[str, list[str]]
    ) -> dict[str, AnnotationInfo]:
        annotations = self._get_multi(annotation_repository, image_id=image_id)
        return {
            annotation.id: AnnotationInfo.model_validate(annotation, from_attributes=True)
            for annotation in annotations
        }

    def get_annotations_by_category_id(
        self, category_id: Union[str, list[str]]
    ) -> dict[str, AnnotationInfo]:
        annotations = self._get_multi(annotation_repository, category_id=category_id)
        return {
            annotation.id: AnnotationInfo.model_validate(annotation, from_attributes=True)
            for annotation in annotations
        }

    # Read custom
    def get_image_count(self, split: SplitType = None, **kwargs) -> int:
        return self._get_count(image_repository, split=split, **kwargs)

    def get_category_count(self) -> int:
        return self._get_count(category_repository)

    def get_annotation_count(self) -> int:
        return self._get_count(annotation_repository)

    def get_image_by_original_file_name(self, original_file_name: str) -> ImageInfo:
        with self.Session() as session:
            image = image_repository.get_multi(
                session, filter_by={"original_file_name": original_file_name}
            )

        if len(image) == 0:
            raise DatabaseNotFoundError(f"image does not exist: {original_file_name}")
        elif len(image) > 1:
            raise DatabaseIntegrityError(f"multiple images exist: {original_file_name}")

        return ImageInfo.model_validate(image[0], from_attributes=True)

    def get_category_by_name(self, name: str) -> CategoryInfo:
        with self.Session() as session:
            category = category_repository.get_multi(session, filter_by={"name": name})

        if len(category) == 0:
            raise DatabaseNotFoundError(f"category does not exist: {name}")
        elif len(category) > 1:
            raise DatabaseIntegrityError(f"multiple categories exist: {name}")

        return CategoryInfo.model_validate(category[0], from_attributes=True)

    # stats
    def get_image_num_by_category_id(
        self, category_id: Union[str, list[str]] = None
    ) -> dict[str, int]:
        annotations = self._get_multi(annotation_repository, category_id=category_id)

        image_num_by_category_id = {}
        for annotation in annotations:
            if annotation.category_id not in image_num_by_category_id:
                image_num_by_category_id[annotation.category_id] = 0
            image_num_by_category_id[annotation.category_id] += 1

        return image_num_by_category_id

    def get_annotation_num_by_category_id(
        self, category_id: Union[str, list[str]] = None
    ) -> dict[str, int]:
        annotations = self._get_multi(annotation_repository, category_id=category_id)

        annotation_num_by_category_id = {}
        for annotation in annotations:
            if annotation.category_id not in annotation_num_by_category_id:
                annotation_num_by_category_id[annotation.category_id] = 0
            annotation_num_by_category_id[annotation.category_id] += 1

        return annotation_num_by_category_id

    def get_annotation_num_by_image_id(
        self, image_id: Union[str, list[str]] = None
    ) -> dict[str, int]:
        annotations = self._get_multi(annotation_repository, image_id=image_id)

        annotation_num_by_image_id = {}
        for annotation in annotations:
            if annotation.image_id not in annotation_num_by_image_id:
                annotation_num_by_image_id[annotation.image_id] = 0
            annotation_num_by_image_id[annotation.image_id] += 1

        return annotation_num_by_image_id

    # Update
    def update_image(self, image_id: str, update_image_info: UpdateImageInfo) -> ImageInfo:
        with self.Session() as session:
            image = image_repository.update(session, image_id, update_image_info)

        return ImageInfo.model_validate(image, from_attributes=True)

    def update_category(
        self, category_id: str, update_category_info: UpdateCategoryInfo
    ) -> CategoryInfo:
        with self.Session() as session:
            category = category_repository.update(session, category_id, update_category_info)

        return CategoryInfo.model_validate(category, from_attributes=True)

    def update_annotation(
        self, annotation_id: str, update_annotation_info: UpdateAnnotationInfo
    ) -> AnnotationInfo:
        with self.Session() as session:
            annotation = annotation_repository.update(session, annotation_id, update_annotation_info)

        return AnnotationInfo.model_validate(annotation, from_attributes=True)

    # Delete
    def delete_image(self, image_id: str) -> None:
        with self.Session() as session:
            image_info = image_repository.remove(session, image_id)

        image_path = self.get_image_path(image_info)
        if Path(image_path).exists():
            io.remove_file(image_path)

    def delete_category(self, category_id: str) -> None:
        with self.Session() as session:
            category_repository.remove(session, category_id)

    def delete_annotation(self, annotation_id: str) -> None:
        with self.Session() as session:
            annotation_repository.remove(session, annotation_id)



from sqlalchemy import JSON, Boolean, Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from .base import Base


class Image(Base):
    __tablename__ = "images"

    id = Column(String, primary_key=True)
    task = Column(String)
    ext = Column(String)
    width = Column(Integer)
    height = Column(Integer)
    original_file_name = Column(String)
    date_captured = Column(String)
    split = Column(String)

    annotations = relationship("Annotation", back_populates="image", cascade="all, delete-orphan")



from typing import Union

from waffle_dough.exception import FieldValidationError
from waffle_dough.type import SplitType


def validate_width(v: int) -> int:
    if v <= 0:
        raise FieldValidationError(f"width must be positive integer: {v}")
    return v


def validate_height(v: int) -> int:
    if v <= 0:
        raise FieldValidationError(f"height must be positive integer: {v}")
    return v


def validate_split(v: Union[str, SplitType]) -> str:
    if v not in list(SplitType):
        raise FieldValidationError(f"split must be one of {list(SplitType)}: {v}")
    return v.lower()



from sqlalchemy.orm import Session

from waffle_dough.database.model import Annotation
from waffle_dough.database.repository.base_repository import CRUDBase
from waffle_dough.database.repository.category_repository import category_repository
from waffle_dough.database.repository.image_repository import image_repository
from waffle_dough.exception.database_exception import *
from waffle_dough.field import AnnotationInfo, UpdateAnnotationInfo


class AnnotationRepository(CRUDBase[Annotation, AnnotationInfo, UpdateAnnotationInfo]):
    def create(self, db: Session, obj_in: AnnotationInfo) -> Annotation:

        if obj_in.category_id is not None:
            category = category_repository.get(db, obj_in.category_id)
            if category is None:
                raise DatabaseNotFoundError(f"Category with id {obj_in.category_id} not found")
            if category.task != obj_in.task:
                raise DatabaseConstraintError(
                    f"Category task {category.task} and annotation task {obj_in.task} mismatch"
                )

        if obj_in.image_id is not None:
            image = image_repository.get(db, obj_in.image_id)
            if image is None:
                raise DatabaseNotFoundError(f"Image with id {obj_in.image_id} not found")

        return super().create(db, obj_in)

    def update(
        self,
        db: Session,
        id: str,
        obj_in: UpdateAnnotationInfo,
    ) -> AnnotationInfo:
        obj = db.get(self.model, id)
        if obj is None:
            raise DatabaseNotFoundError(f"Object with id {id} not found")
        obj_in_data = {col.name: str(getattr(obj, col.name)) for col in obj.__table__.columns}
        if isinstance(obj_in, dict):
            update_data = obj_in
        else:
            update_data = obj_in.model_dump(exclude_unset=True)
        for field in obj_in_data:
            if field in update_data and update_data[field] is not None:
                setattr(obj, field, update_data[field])

        if obj_in.category_id is not None:
            category = category_repository.get(db, obj_in.category_id)
            if category is None:
                raise DatabaseNotFoundError(f"Category with id {obj_in.category_id} not found")

        db.add(obj)
        self.commit(db)
        db.refresh(obj)
        return obj


annotation_repository = AnnotationRepository(Annotation)



from waffle_utils.enum import StrEnum as BaseType


class SplitType(BaseType):
    TRAIN = "train"
    VALIDATION = "validation"
    TEST = "test"
    BACKGROUND = "background"
    UNLABELED = "unlabeled"
    PREDICTION = "prediction"
    UNSET = "unset"



from .progress.progress import DatasetAdapterProgressCallback
from .progress.tqdm import DatasetAdapterTqdmProgressCallback

__all__ = [
    "DatasetAdapterProgressCallback",
    "DatasetAdapterTqdmProgressCallback",
]



from waffle_utils.hook import BaseHook


class BaseAdapter(BaseHook):
    def on_loop_start(self, total_steps: int):
        pass

    def on_loop_end(self):
        pass

    def on_step_start(self):
        pass

    def on_step_end(self, current_step: int):
        pass



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import StaticPool

from .model import Base


def create_session(db_path: str = None) -> sessionmaker:
    """Create a session.

    Args:
        db_path (str, optional): Path to the database. Defaults to ":memory:".

    Returns:
        sessionmaker: sessionmaker.
    """
    engine = create_engine(
        "sqlite://" + ("/" + db_path if db_path is not None else ""),
        connect_args={"check_same_thread": False},
        poolclass=StaticPool,
    )
    Base.metadata.create_all(engine)
    return sessionmaker(bind=engine)



from waffle_dough.database.model import Category
from waffle_dough.database.repository.base_repository import CRUDBase
from waffle_dough.field import CategoryInfo, UpdateCategoryInfo


class CategoryRepository(CRUDBase[Category, CategoryInfo, UpdateCategoryInfo]):
    pass


category_repository = CategoryRepository(Category)



import time

from ..adapter_callback import BaseDatasetAdapterCallback


class DatasetAdapterProgressCallback(BaseDatasetAdapterCallback):
    _total_steps = None
    _current_step = None
    _last_updated_step = None
    _last_updated_time = None
    _start_time = None
    _started = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def total_steps(self) -> int:
        return DatasetAdapterProgressCallback._total_steps

    @total_steps.setter
    def total_steps(self, total_steps: int):
        DatasetAdapterProgressCallback._total_steps = total_steps

    @property
    def current_step(self) -> int:
        return DatasetAdapterProgressCallback._current_step

    @current_step.setter
    def current_step(self, current_step: int):
        DatasetAdapterProgressCallback._current_step = current_step
        self.last_updated_step = current_step
        self.last_updated_time = time.time()

    @property
    def last_updated_step(self) -> int:
        return DatasetAdapterProgressCallback._last_updated_step

    @last_updated_step.setter
    def last_updated_step(self, last_updated_step: int):
        DatasetAdapterProgressCallback._last_updated_step = last_updated_step

    @property
    def last_updated_time(self) -> float:
        return DatasetAdapterProgressCallback._last_updated_time

    @last_updated_time.setter
    def last_updated_time(self, last_updated_time: float):
        DatasetAdapterProgressCallback._last_updated_time = last_updated_time

    @property
    def start_time(self) -> float:
        return DatasetAdapterProgressCallback._start_time

    @start_time.setter
    def start_time(self, start_time: float):
        DatasetAdapterProgressCallback._start_time = start_time

    @property
    def started(self) -> bool:
        return DatasetAdapterProgressCallback._started

    @started.setter
    def started(self, started: bool):
        DatasetAdapterProgressCallback._started = started

    def get_remaining_time(self) -> float:
        if (
            self.start_time is None
            or self.current_step is None
            or self.current_step == 0
            or self.total_steps is None
        ):
            return float("inf")

        elapsed_time = self.last_updated_time - self.start_time
        return elapsed_time * (self.total_steps - self.current_step) / self.current_step

    def initialize(self):
        self.total_steps = None
        self.current_step = None
        self.start_time = None

        self.started = False

    def on_loop_start(self, total_steps: int):
        if self.started:
            if self.total_steps != total_steps:
                raise ValueError("total_steps must be same with previous one")
            return
        self.started = True
        self.total_steps = total_steps
        self.current_step = 0
        self.start_time = time.time()

    def on_loop_end(self):
        self.current_step = self.total_steps
        self.started = False

    def on_step_start(self):
        pass

    def on_step_end(self, current_step: int):
        if current_step == self.last_updated_step:
            return
        if current_step > self.total_steps or current_step < self.current_step or current_step < 0:
            raise ValueError("current_step must be between 0 and total_steps")
        self.current_step = current_step



"""
Waffle Dough Dataset

This module contains the Waffle Dough Dataset class.


Example:
    >>> from waffle_dough.dataset import WaffleDataset
    >>> dataset = WaffleDataset.new(
    ...     name="my_dataset",
    ...     task="classification",
    ...     root_dir="~/datasets",
    ... )
    >>> dataset.add_category(
    ...     CategoryInfo.classification(
    ...         name="cat",
    ...         label=0,

    )

"""
import logging
import os
import random
from dataclasses import asdict, dataclass
from functools import cached_property
from pathlib import Path
from typing import Union

from tqdm import tqdm
from waffle_utils.file import io
from waffle_utils.logger import datetime_now, initialize_logger

from waffle_dough.database.service import DatabaseService
from waffle_dough.dataset.adapter import CocoAdapter
from waffle_dough.dataset.util.iterator import Iterator
from waffle_dough.dataset.util.visualize import visualize
from waffle_dough.exception.base_exception import BaseException
from waffle_dough.exception.dataset_exception import *
from waffle_dough.field import (
    AnnotationInfo,
    CategoryInfo,
    ImageInfo,
    UpdateAnnotationInfo,
    UpdateCategoryInfo,
    UpdateImageInfo,
)
from waffle_dough.image import io as image_io
from waffle_dough.type import SplitType, TaskType

initialize_logger(
    console_level=os.environ.get("WAFFLE_DATASET_CONSOLE_LOG_LEVEL", "INFO"),
    file_level=os.environ.get("WAFFLE_DATASET_FILE_LOG_LEVEL", "INFO"),
    root_level=os.environ.get("WAFFLE_DATASET_ROOT_LOG_LEVEL", "INFO"),
)
logger = logging.getLogger(__name__)


@dataclass
class DatasetInfo:
    name: str
    task: str
    categories: list[dict]
    created_at: str
    updated_at: str

    def to_dict(self):
        return asdict(self)

    @staticmethod
    def from_dict(dataset_info: dict):
        return DatasetInfo(
            name=dataset_info["name"],
            task=dataset_info["task"],
            categories=dataset_info["categories"],
            created_at=dataset_info["created_at"],
            updated_at=dataset_info["updated_at"],
        )


class WaffleDataset:
    DATASET_INFO_FILE_NAME = "dataset.yaml"
    DATABASE_FILE_NAME = "database.sqlite3"
    IMAGE_DIR_NAME = "images"

    def __init__(
        self,
        name: str,
        task: Union[str, TaskType] = None,
        root_dir: Union[str, Path] = None,
    ):
        self.name = name
        self.root_dir = root_dir

        if self.initialized():
            dataset_info = self.get_dataset_info()
            if task is not None and task.lower() != dataset_info.task:
                raise DatasetTaskError(f"Invalid task: {task}")
            self.task = dataset_info.task
        else:
            if task is None:
                raise DatasetTaskError(f"Task is not specified")
            self.task = task
            self.initialize()

    def __repr__(self) -> str:
        return f"WaffleDataset(name={self.name}, task={self.task}, root_dir={self.root_dir})"

    def __str__(self) -> str:
        return f"WaffleDataset(name={self.name}, task={self.task}, root_dir={self.root_dir})"

    def initialize(self):
        io.make_directory(self.dataset_dir)
        io.make_directory(self.image_dir)
        self.create_dataset_info()

    def initialized(self) -> bool:
        return self.dataset_info_file.exists()

    def create_dataset_info(self) -> DatasetInfo:
        dataset_info = DatasetInfo(
            name=self.name,
            task=self.task,
            categories=[category.to_dict() for category in self.categories],
            created_at=datetime_now(),
            updated_at=datetime_now(),
        )
        io.save_yaml(dataset_info.to_dict(), self.dataset_info_file)
        return dataset_info

    def get_dataset_info(self) -> DatasetInfo:
        return DatasetInfo.from_dict(io.load_yaml(self.dataset_info_file))

    def update_dataset_info(self) -> DatasetInfo:
        dataset_info = self.get_dataset_info()
        dataset_info.categories = [category.to_dict() for category in self.categories]
        dataset_info.updated_at = datetime_now()
        io.save_yaml(dataset_info.to_dict(), self.dataset_info_file)
        return dataset_info

    # decorators
    def update_dataset_decorator(func):
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            self.update_dataset_info()
            return result

        return wrapper

    def exception_decorator(func):
        def wrapper(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except BaseException as e:
                raise e
            except Exception as e:
                exc = BaseException(e)
                exc.__traceback__ = e.__traceback__
                raise exc

        return wrapper

    # properties
    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = str(name)

    @property
    def task(self) -> str:
        return self._task

    @task.setter
    def task(self, task: Union[str, TaskType]):
        if task not in list(TaskType):
            raise DatasetTaskError(f"Invalid task: {task}")
        self._task = task.lower()

    @property
    def root_dir(self) -> Path:
        return self._root_dir

    @root_dir.setter
    def root_dir(self, root_dir: str):
        self._root_dir = self.parse_root_dir(root_dir)

    @property
    def dataset_dir(self) -> Path:
        return self.root_dir / self.name

    @property
    def dataset_info_file(self) -> Path:
        return self.dataset_dir / self.DATASET_INFO_FILE_NAME

    @property
    def image_dir(self) -> Path:
        return self.dataset_dir / self.IMAGE_DIR_NAME

    @property
    def database_file(self) -> Path:
        return self.dataset_dir / self.DATABASE_FILE_NAME

    @property
    def log_file(self) -> Path:
        return self.dataset_dir / "logs" / "dataset.log"

    @cached_property
    def database_service(self) -> DatabaseService:
        return DatabaseService(str(self.database_file), image_directory=self.image_dir)

    @property
    def category_dict(self) -> dict[str, CategoryInfo]:
        return self.database_service.get_categories()

    @property
    def categories(self) -> list[CategoryInfo]:
        return list(self.category_dict.values())

    @property
    def category_names(self) -> list[str]:
        return [category.name for category in self.categories]

    @property
    def image_dict(self) -> dict[str, ImageInfo]:
        return self.database_service.get_images()

    @property
    def images(self) -> list[ImageInfo]:
        return list(self.image_dict.values())

    @property
    def annotation_dict(self) -> dict[str, AnnotationInfo]:
        return self.database_service.get_annotations()

    @property
    def annotations(self) -> list[AnnotationInfo]:
        return list(self.database_service.get_annotations().values())

    # methods (CRUD)
    @update_dataset_decorator
    @exception_decorator
    def add_category(
        self, category_info: Union[CategoryInfo, list[CategoryInfo], dict, list[dict]]
    ) -> list[CategoryInfo]:
        category_infos = []
        for c in category_info if isinstance(category_info, list) else [category_info]:
            if isinstance(c, dict):
                c = CategoryInfo.from_dict(task=self.task, d=c)
            if c.task != self.task:
                raise DatasetTaskError(f"Invalid task: {c.task}")
            category_infos.append(c)

        for category_info in category_infos:
            self.database_service.add_category(category_info)

        return category_infos

    @update_dataset_decorator
    @exception_decorator
    def add_image(
        self,
        image: Union[str, Path],
        image_info: Union[ImageInfo, list[ImageInfo], dict, list[dict]] = None,
    ) -> list[ImageInfo]:
        images = image if isinstance(image, list) else [image]
        if image_info is None:
            image_infos = []
            for image in images:
                img = image_io.cv2_imread(image)
                image_infos.append(
                    ImageInfo(
                        ext=Path(image).suffix,
                        width=img.shape[1],
                        height=img.shape[0],
                        original_file_name=str(image),
                        task=TaskType.AGNOSTIC,
                    )
                )
        else:
            image_infos = []
            for image_info in image_info if isinstance(image_info, list) else [image_info]:
                if isinstance(image_info, dict):
                    image_info = ImageInfo.from_dict(task=TaskType.AGNOSTIC, d=image_info)
                if image_info.task != TaskType.AGNOSTIC:
                    raise DatasetTaskError(f"Invalid task: {image_info.task}")
                image_infos.append(image_info)

            if len(images) != len(image_infos):
                raise DatasetImportError(f"Number of images and image_infos must be same")

        for image, image_info in zip(images, image_infos):
            if isinstance(image_info, dict):
                image_info = ImageInfo.from_dict(task=TaskType.AGNOSTIC, d=image_info)
            self.database_service.add_image(image, image_info)

        return image_infos

    @update_dataset_decorator
    @exception_decorator
    def add_annotation(
        self, annotation_info: Union[AnnotationInfo, list[AnnotationInfo], dict, list[dict]]
    ) -> list[AnnotationInfo]:
        annotation_infos = []
        for a in annotation_info if isinstance(annotation_info, list) else [annotation_info]:
            if isinstance(a, dict):
                a = AnnotationInfo.from_dict(task=self.task, d=a)
            if a.task != self.task:
                raise DatasetTaskError(f"Invalid task: {a.task}")
            annotation_infos.append(a)

        for annotation_info in annotation_infos:
            self.database_service.add_annotation(annotation_info)

        return annotation_infos

    @exception_decorator
    def get_image_dict(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split: Union[str, SplitType] = None,
    ) -> dict[str, ImageInfo]:
        if category_id is None:
            images = self.database_service.get_images(image_id=image_id, split=split)
        else:
            images = self.database_service.get_images_by_category_id(
                category_id=category_id, split=split
            )
        return images

    @exception_decorator
    def get_images(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split: Union[str, SplitType] = None,
    ) -> list[ImageInfo]:
        return list(
            self.get_image_dict(image_id=image_id, category_id=category_id, split=split).values()
        )

    @exception_decorator
    def get_image_path(self, image: ImageInfo):
        return self.database_service.get_image_path(image)

    @exception_decorator
    def get_annotation_dict(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split: Union[str, SplitType] = None,
    ) -> dict[str, AnnotationInfo]:
        images = self.get_images(image_id=image_id, category_id=category_id, split=split)
        annotations = self.database_service.get_annotations_by_image_id(
            image_id=[image.id for image in images]
        )
        return annotations

    @exception_decorator
    def get_annotations(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split: Union[str, SplitType] = None,
    ) -> list[AnnotationInfo]:
        return list(
            self.get_annotation_dict(
                image_id=image_id, category_id=category_id, split=split
            ).values()
        )

    @exception_decorator
    def get_category_dict(
        self, category_id: Union[str, list[str]] = None
    ) -> dict[str, CategoryInfo]:
        return self.database_service.get_categories(category_id=category_id)

    @exception_decorator
    def get_categories(self, category_id: Union[str, list[str]] = None) -> list[CategoryInfo]:
        return list(self.get_category_dict(category_id=category_id).values())

    @exception_decorator
    def get_mapper(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split: Union[str, SplitType] = None,
        labeled_only: bool = False,
    ) -> dict[str, dict]:
        image_dict = self.get_image_dict(image_id=image_id, category_id=category_id, split=split)
        category_dict = self.get_category_dict(category_id=category_id)

        mapper = {}
        for image_id, image in image_dict.items():
            annotations = self.get_annotations(image_id=image_id)
            categories = [category_dict[annotation.category_id] for annotation in annotations]
            if labeled_only and len(categories) == 0:
                continue
            mapper[image_id] = {
                "image_path": self.get_image_path(image),
                "image_info": image,
                "annotations": annotations,
                "categories": categories,
            }

        return mapper

    @update_dataset_decorator
    @exception_decorator
    def update_image(self, image_id: str, update_image_info: UpdateImageInfo) -> ImageInfo:
        return self.database_service.update_image(image_id, update_image_info)

    @update_dataset_decorator
    @exception_decorator
    def update_category(
        self, category_id: str, update_category_info: UpdateCategoryInfo
    ) -> CategoryInfo:
        return self.database_service.update_category(category_id, update_category_info)

    @update_dataset_decorator
    @exception_decorator
    def update_annotation(
        self, annotation_id: str, update_annotation_info: UpdateAnnotationInfo
    ) -> AnnotationInfo:
        return self.database_service.update_annotation(annotation_id, update_annotation_info)

    @update_dataset_decorator
    @exception_decorator
    def delete_image(self, image_id: str):
        self.database_service.delete_image(image_id)

    @update_dataset_decorator
    @exception_decorator
    def delete_category(self, category_id: str):
        self.database_service.delete_category(category_id)

    @update_dataset_decorator
    @exception_decorator
    def delete_annotation(self, annotation_id: str):
        self.database_service.delete_annotation(annotation_id)

    # methods (class methods)
    @classmethod
    @exception_decorator
    def get_dataset_list(cls, task: Union[str, TaskType] = None, root_dir: str = None) -> list[str]:
        root_dir = cls.parse_root_dir(root_dir)

        dataset_list = []
        if not root_dir.exists():
            return dataset_list

        for dataset_dir in Path(root_dir).iterdir():
            if not dataset_dir.is_dir():
                continue
            dataset_info_file = dataset_dir / cls.DATASET_INFO_FILE_NAME
            if dataset_info_file.exists():
                dataset_info = DatasetInfo.from_dict(io.load_yaml(dataset_info_file))
                if task is None or TaskType.from_str(task) == dataset_info.task:
                    dataset_list.append(Path(dataset_dir).name)
        return dataset_list

    @classmethod
    @exception_decorator
    def parse_root_dir(cls, root_dir: str = None) -> Path:
        if root_dir is None:
            root_dir = os.environ.get("WAFFLE_DATASET_ROOT_DIR", "datasets")
        return Path(root_dir).absolute()

    @classmethod
    @exception_decorator
    def new(
        cls,
        name: str,
        task: Union[str, TaskType],
        root_dir: Union[str, Path] = None,
    ) -> "WaffleDataset":
        if name in WaffleDataset.get_dataset_list(root_dir=root_dir):
            raise DatasetAlreadyExistsError(f"Dataset '{name}' already exists")
        dataset = WaffleDataset(name, task, root_dir=root_dir)
        logger.info(f"Dataset created [{name}]\n{dataset}")
        return dataset

    @classmethod
    @exception_decorator
    def load(
        cls,
        name: str,
        root_dir: Union[str, Path] = None,
    ) -> "WaffleDataset":
        if name not in WaffleDataset.get_dataset_list(root_dir=root_dir):
            raise DatasetNotFoundError(f"Dataset '{name}' does not exists")
        dataset = WaffleDataset(name, root_dir=root_dir)
        logger.info(f"Dataset loaded [{name}]\n{dataset}")
        return dataset

    @classmethod
    @exception_decorator
    def delete(
        cls,
        name: str,
        root_dir: Union[str, Path] = None,
    ):
        dataset = WaffleDataset.load(name, root_dir=root_dir)
        io.remove_directory(dataset.dataset_dir, recursive=True)
        logger.info(f"Dataset deleted [{name}]\n{dataset}")

    @classmethod
    @exception_decorator
    def copy(
        cls,
        src_name: str,
        dst_name: str,
        root_dir: Union[str, Path] = None,
    ) -> "WaffleDataset":
        src_dataset = WaffleDataset.load(src_name, root_dir=root_dir)
        dst_dataset = WaffleDataset.new(dst_name, src_dataset.task, root_dir=root_dir)

        try:
            io.copy_files_to_directory(src_dataset.image_dir, dst_dataset.image_dir)
        except FileNotFoundError:
            pass
        io.copy_file(src_dataset.database_file, dst_dataset.database_file)
        dst_dataset.update_dataset_info()
        logger.info(f"Dataset copied [{src_name} -> {dst_name}]\n{dst_dataset}")

        return dst_dataset

    @classmethod
    def from_coco(
        cls,
        name: str,
        task: Union[str, TaskType],
        coco: Union[str, Path, dict],
        coco_image_dir: Union[str, Path],
        root_dir: Union[str, Path] = None,
    ) -> "WaffleDataset":
        dataset = WaffleDataset.new(name, task, root_dir=root_dir)

        try:
            adapter = CocoAdapter.from_target(coco, task=task)

            dataset.add_category(list(adapter.categories.values()))

            image_path = []
            image_infos = []
            for image in adapter.images.values():
                image_path.append(Path(coco_image_dir, image.original_file_name))
                image_infos.append(image)
            dataset.add_image(image_path, image_infos)
            dataset.add_annotation(list(adapter.annotations.values()))

            return dataset

        except Exception as e:
            WaffleDataset.delete(name, root_dir=root_dir)
            raise e

    @exception_decorator
    def random_split(
        self,
        train_ratio: float = 0.8,
        val_ratio: float = 0.1,
        test_ratio: float = 0.1,
        seed: int = 42,
    ):
        image_ids = list(self.get_mapper(labeled_only=True).keys())
        if len(image_ids) == 0:
            raise DatasetEmptyError(f"There are no labeled images")

        random.seed(seed)
        random.shuffle(image_ids)

        # make sure that the ratios are valid
        train_ratio = float(train_ratio) if train_ratio else 0.0
        val_ratio = float(val_ratio) if val_ratio else 0.0
        test_ratio = float(test_ratio) if test_ratio else 0.0
        if (
            any([ratio < 0 for ratio in [train_ratio, val_ratio, test_ratio]])
            or sum([train_ratio, val_ratio, test_ratio]) == 0
        ):
            raise DatasetSplitError(
                "Ratio must be non-negative float or int, and their sum must be positive"
            )

        # initialize split
        for image_id in (
            list(self.get_image_dict(split=SplitType.TRAIN).keys())
            + list(self.get_image_dict(split=SplitType.VALIDATION).keys())
            + list(self.get_image_dict(split=SplitType.TEST).keys())
        ):
            self.update_image(image_id, UpdateImageInfo(split=SplitType.UNSET))

        # make the sum of ratios to 1
        ratio_sum = train_ratio + val_ratio + test_ratio
        train_ratio = train_ratio / ratio_sum
        val_ratio = val_ratio / ratio_sum
        test_ratio = test_ratio / ratio_sum

        train_image_ids = image_ids[: int(len(image_ids) * train_ratio)]
        val_image_ids = image_ids[
            int(len(image_ids) * train_ratio) : int(len(image_ids) * (train_ratio + val_ratio))
        ]
        test_image_ids = image_ids[int(len(image_ids) * (train_ratio + val_ratio)) :]

        for image_id in train_image_ids:
            self.update_image(image_id, UpdateImageInfo(split=SplitType.TRAIN))
        for image_id in val_image_ids:
            self.update_image(image_id, UpdateImageInfo(split=SplitType.VALIDATION))
        for image_id in test_image_ids:
            self.update_image(image_id, UpdateImageInfo(split=SplitType.TEST))

        logger.info(
            f"Dataset splitted [{self.name}]\ntrain: {len(train_image_ids)}, val: {len(val_image_ids)}, test: {len(test_image_ids)}"
        )

    @exception_decorator
    def get_dataset_iterator(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split: Union[str, SplitType] = None,
        preprocess=None,
        augment=None,
    ) -> Iterator:
        return Iterator(
            self.get_mapper(image_id=image_id, category_id=category_id, split=split),
            preprocess=preprocess,
            augment=augment,
        )

    @exception_decorator
    def visualize(
        self,
        image_id: Union[str, list[str]] = None,
        category_id: Union[str, list[str]] = None,
        split: Union[str, SplitType] = None,
        result_dir: Union[str, Path] = None,
        show: bool = False,
    ) -> Path:
        it = self.get_dataset_iterator(image_id=image_id, category_id=category_id, split=split)
        category_dict = self.get_category_dict()

        result_dir = Path(result_dir or self.dataset_dir / "visualized").absolute()
        logger.info(f"Visualizing dataset [{self.name}] to {result_dir}")

        pgbar = tqdm(it, total=len(it), desc="Visualizing dataset")
        for data in pgbar:
            draw = visualize(
                image=data.image,
                annotations=data.annotations,
                category_dict=category_dict,
                image_info=data.image_info,
            )

            image_path = result_dir / (data.image_info.id + data.image_info.ext)
            image_io.cv2_imwrite(image_path, draw, create_directory=True)

            if show:
                image_io.cv2_imshow("vis", draw)

        logger.info(f"Visualizing dataset [{self.name}] to {result_dir} finished")

        return Path(result_dir)

    # @exception_decorator
    # def experimental_filter(
    #     self,
    #     new_dataset_name: str,
    #     image_lambda: callable[ImageInfo, bool] = None,
    #     annotation_lambda: callable[AnnotationInfo, bool] = None,
    #     category_lambda: callable[CategoryInfo, bool] = None,
    #     root_dir: Union[str, Path] = None,
    # ) -> "WaffleDataset":
    #     pass

    # @exception_decorator
    # def experimental_manipulate(
    #     self,
    #     new_dataset_name: str,
    #     task: Union[str, TaskType],
    #     image_lambda: callable[ImageInfo, bool] = None,
    #     annotation_lambda: callable[AnnotationInfo, bool] = None,
    #     category_lambda: callable[CategoryInfo, bool] = None,
    #     root_dir: Union[str, Path] = None,
    # ) -> "WaffleDataset":
    #     pass



from waffle_utils.enum import StrEnum as BaseType


class TaskType(BaseType):
    CLASSIFICATION = "classification"
    OBJECT_DETECTION = "object_detection"
    SEMANTIC_SEGMENTATION = "semantic_segmentation"
    INSTANCE_SEGMENTATION = "instance_segmentation"
    KEYPOINT_DETECTION = "keypoint_detection"
    TEXT_RECOGNITION = "text_recognition"
    REGRESSION = "regression"

    AGNOSTIC = "agnostic"



from typing import Union

import numpy as np

from waffle_dough.exception import FieldValidationError


def validate_bbox(v: list[float]) -> list[float]:
    if v:
        if len(v) != 4:
            raise FieldValidationError("the length of bbox should be 4.")
        if not all([isinstance(v_, (float)) for v_ in v]):
            raise FieldValidationError("the bbox values should be float.")
        if not all([0 <= v_ < 1 for v_ in v[:2]]):
            raise FieldValidationError("the [left, top] should be in [0, 1).")
        if not all([0 < v_ <= 1 for v_ in v[2:]]):
            raise FieldValidationError("the [width, height] should be in (0, 1].")
    return v


def validate_segmentation(
    v: Union[dict, list[list[float]], np.ndarray],
) -> Union[list[list[float]], np.ndarray]:
    if v:
        if not isinstance(v, (list, np.ndarray)):
            raise FieldValidationError("segmentation should be list or np.ndarray.")

        for segment in v:
            if len(segment) % 2 != 0:
                raise FieldValidationError("the length of segmentation should be divisible by 2.")
            if len(segment) < 6:
                raise FieldValidationError(
                    "the length of segmentation should be greater than or equal to 6."
                )
    return v


def validate_area(v: float) -> float:
    if v:
        if v < 0:
            raise FieldValidationError("area should be greater than or equal to 0.")
    return v


def validate_keypoints(v: list[float]) -> list[float]:
    if v:
        if len(v) % 3 != 0:
            raise FieldValidationError("the length of keypoints should be divisible by 3.")
    return v


def validate_num_keypoints(v: int) -> int:
    if v:
        if v < 0:
            raise FieldValidationError("num_keypoints should be greater than or equal to 0.")
    return v


def validate_caption(v: str) -> str:
    if v:
        if not isinstance(v, str):
            raise FieldValidationError("caption should be str.")
    return v


def validate_value(v: Union[int, float]) -> float:
    if v:
        if not isinstance(v, (int, float)):
            raise FieldValidationError("value should be int or float.")
        v = float(v)
    return v


def validate_iscrowd(v: int) -> int:
    if v:
        if v not in [0, 1]:
            raise FieldValidationError("iscrowd should be 0 or 1.")
    return v


def validate_score(v: float) -> float:
    if v:
        if v < 0 or v > 1:
            raise FieldValidationError("score should be in [0, 1].")
    return v


def validate_is_prediction(v: bool) -> bool:
    if v:
        if not isinstance(v, bool):
            raise FieldValidationError("is_prediction should be bool.")
    return v



from typing import Union

import numpy as np

from waffle_dough.field import AnnotationInfo, CategoryInfo, ImageInfo
from waffle_dough.image.io import cv2_imread
from waffle_dough.type import ColorType


class Data:
    def __init__(
        self,
        image: np.ndarray,
        image_path: str,
        image_info: ImageInfo,
        annotations: list[AnnotationInfo],
        categories: list[CategoryInfo],
    ):
        self.image = image
        self.image_path = image_path
        self.image_info = image_info
        self.annotations = annotations
        self.categories = categories

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Data(image={self.image.shape}, image_path={self.image_path}, image_info={self.image_info}, annotations={self.annotations}, categories={self.categories})"


class Iterator:
    def __init__(
        self,
        mapper: dict[dict],
        preprocess=None,
        augment=None,
        color_type: Union[str, ColorType] = ColorType.RGB,
    ):
        self.mapper = mapper
        self.image_ids = list(mapper.keys())
        self.preprocess = preprocess
        self.augment = augment
        self.color_type = color_type

    def __len__(self):
        return len(self.image_ids)

    def __getitem__(self, idx) -> Data:
        image_id = self.image_ids[idx]

        image_path = self.mapper[image_id]["image_path"]
        image = cv2_imread(image_path, color_type=self.color_type)
        if self.preprocess is not None:
            image = self.preprocess(image)
        if self.augment is not None:
            image = self.augment(image)

        image_info: ImageInfo = self.mapper[image_id]["image_info"]
        annotations: list[AnnotationInfo] = self.mapper[image_id]["annotations"]
        categories: list[CategoryInfo] = self.mapper[image_id]["categories"]

        return Data(image, image_path, image_info, annotations, categories)



from typing import Union

from waffle_dough.type.annotation_type import BoxType


def x1y1x2y2_to_x1y1wh(box: list[float]) -> list[float]:
    x1, y1, x2, y2 = box
    return [x1, y1, x2 - x1, y2 - y1]


def x1y1x2y2_to_cxcywh(box: list[float]) -> list[float]:
    x1, y1, x2, y2 = box
    return [(x1 + x2) / 2, (y1 + y2) / 2, x2 - x1, y2 - y1]


def x1y1wh_to_x1y1x2y2(box: list[float]) -> list[float]:
    x1, y1, w, h = box
    return [x1, y1, x1 + w, y1 + h]


def x1y1wh_to_cxcywh(box: list[float]) -> list[float]:
    x1, y1, w, h = box
    return [x1 + w / 2, y1 + h / 2, w, h]


def cxcywh_to_x1y1x2y2(box: list[float]) -> list[float]:
    cx, cy, w, h = box
    return [cx - w / 2, cy - h / 2, cx + w / 2, cy + h / 2]


def cxcywh_to_x1y1wh(box: list[float]) -> list[float]:
    cx, cy, w, h = box
    return [cx - w / 2, cy - h / 2, w, h]


def convert_box(
    box: list[float], src_type: Union[str, BoxType], dst_type: Union[str, BoxType]
) -> list[float]:
    """Convert box from src_type to dst_type.

    Args:
        box (list[float]): box
        src_type (Union[str, BoxType]): source type
        dst_type (Union[str, BoxType]): destination type

    Returns:
        list[float]: converted bbox
    """

    if src_type == dst_type:
        return box

    if src_type == BoxType.XYXY:
        if dst_type == BoxType.XYWH:
            return x1y1x2y2_to_x1y1wh(box)
        elif dst_type == BoxType.CXCYWH:
            return x1y1x2y2_to_cxcywh(box)
    elif src_type == BoxType.XYWH:
        if dst_type == BoxType.XYXY:
            return x1y1wh_to_x1y1x2y2(box)
        elif dst_type == BoxType.CXCYWH:
            return x1y1wh_to_cxcywh(box)
    elif src_type == BoxType.CXCYWH:
        if dst_type == BoxType.XYXY:
            return cxcywh_to_x1y1x2y2(box)
        elif dst_type == BoxType.XYWH:
            return cxcywh_to_x1y1wh(box)
    else:
        raise ValueError(f"Cannot convert bbox from {src_type} to {dst_type}")


def get_box_area(box: list[float], box_type: Union[str, BoxType]) -> float:
    """Get area from box.

    Args:
        box (list[float]): box
        box_type (Union[str, BoxType]): box type

    Returns:
        float: area
    """

    _, _, w, h = convert_box(box, box_type, BoxType.XYWH)
    return w * h



from .progress import DatasetAdapterProgressCallback
from .tqdm import DatasetAdapterTqdmProgressCallback

__all__ = [
    "DatasetAdapterProgressCallback",
    "DatasetAdapterTqdmProgressCallback",
]



import logging
import os

import dotenv
from waffle_utils.file import io

dotenv.load_dotenv()


class Config:
    # Dataset Configuration
    DATASET_ROOT_DIR: str = os.environ.get("WAFFLE_DATASET_ROOT_DIR", "./datasets")


logger = logging.getLogger(__name__)

io.make_directory(Config.DATASET_ROOT_DIR)
logger.info(f"Dataset root directory: {Config.DATASET_ROOT_DIR}")



import tqdm

from . import DatasetAdapterProgressCallback


class DatasetAdapterTqdmProgressCallback(DatasetAdapterProgressCallback):
    def __init__(self, desc: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.desc = desc
        self.tqdm_bar = None

    @property
    def desc(self) -> str:
        return self._desc

    @desc.setter
    def desc(self, desc: str):
        self._desc = desc or ""

    @property
    def tqdm_bar(self) -> tqdm.tqdm:
        return self._tqdm_bar

    @tqdm_bar.setter
    def tqdm_bar(self, tqdm_bar: tqdm.tqdm):
        self._tqdm_bar = tqdm_bar

    def on_loop_start(self, total_steps: int):
        super().on_loop_start(total_steps)
        self.tqdm_bar = tqdm.tqdm(total=total_steps, desc=self.desc)

    def on_loop_end(self):
        super().on_loop_end()
        self.tqdm_bar.close()
        self.tqdm_bar = None

    def on_step_end(self, current_step: int):
        super().on_step_end(current_step)
        self.tqdm_bar.update(self.current_step - self.tqdm_bar.n)



from sqlalchemy import JSON, Boolean, Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from .base import Base


class Category(Base):
    __tablename__ = "categories"

    id = Column(String, primary_key=True)
    task = Column(String)
    name = Column(String, unique=True)
    supercategory = Column(String)
    keypoints = Column(JSON)
    skeleton = Column(JSON)

    annotations = relationship("Annotation", back_populates="category", cascade="all, delete-orphan")



from pathlib import Path
from typing import Union

import cv2
import numpy as np
from waffle_utils.file import io

from waffle_dough.type import ColorType

COLOR_TYPE_MAP = {
    ColorType.GRAY: "GRAY",
    ColorType.RGB: "RGB",
    ColorType.BGR: "BGR",
}


def cv2_cvt_color(
    image: np.ndarray, src_type: Union[str, ColorType], dst_type: Union[str, ColorType]
) -> np.ndarray:
    if src_type == dst_type:
        return image

    src_type = COLOR_TYPE_MAP.get(src_type)
    dst_type = COLOR_TYPE_MAP.get(dst_type)

    return cv2.cvtColor(image, getattr(cv2, f"COLOR_{src_type}2{dst_type}"))


def cv2_imread(
    path: Union[str, Path], color_type: Union[str, ColorType] = ColorType.BGR
) -> np.ndarray:
    image = np.fromfile(str(path), dtype=np.uint8)
    image = cv2.imdecode(image, cv2.IMREAD_UNCHANGED)
    image = cv2_cvt_color(image, ColorType.BGR, color_type)
    return image


def cv2_imwrite(
    path: Union[str, Path],
    image: np.ndarray,
    create_directory: bool = False,
    color_type: Union[str, ColorType] = ColorType.BGR,
):
    output_path = Path(path)
    if create_directory:
        io.make_directory(output_path.parent)

    save_type = output_path.suffix
    bgr_image = cv2_cvt_color(image, color_type, ColorType.BGR)
    ret, img_arr = cv2.imencode(save_type, bgr_image)
    if ret:
        with open(str(output_path), mode="w+b") as f:
            img_arr.tofile(f)
    else:
        raise ValueError(f"Failed to save image: {path}")


def cv2_imshow(
    window_name: str,
    image: np.ndarray,
    delay: int = 0,
):
    cv2.imshow(window_name, image)
    cv2.waitKey(delay)
    cv2.destroyAllWindows()



# Waffle Dough (based on Ultralytics YOLO 🚀, GPL-3.0 license)

import re
from pathlib import Path

import pkg_resources as pkg
from setuptools import find_packages, setup

# Settings
FILE = Path(__file__).resolve()
PARENT = FILE.parent  # root directory
README = (PARENT / "README.md").read_text(encoding="utf-8")
REQUIREMENTS = [
    f"{x.name}{x.specifier}"
    for x in pkg.parse_requirements((PARENT / "requirements.txt").read_text())
]
PKG_REQUIREMENTS = ["sentry_sdk"]  # pip-only requirements


def get_version():
    file = PARENT / "waffle_dough/__init__.py"
    return re.search(
        r'^__version__ = [\'"]([^\'"]*)[\'"]',
        file.read_text(encoding="utf-8"),
        re.M,
    )[1]


setup(
    name="waffle_dough",  # name of pypi package
    version=get_version(),  # version of pypi package
    python_requires=">=3.9",
    license="GPL-3.0",
    description="Waffle Dough",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/snuailab/waffle_dough",
    project_urls={
        "Bug Reports": "https://github.com/snuailab/waffle_dough/issues",
        "Source": "https://github.com/snuailab/waffle_dough",
    },
    author="SNUAILAB",
    author_email="huijae.lee@snuailab.ai",
    packages=find_packages(),  # required
    include_package_data=True,
    install_requires=REQUIREMENTS + PKG_REQUIREMENTS,
    # extras_require={
    #     'dev': ['check-manifest', 'pytest', 'pytest-cov', 'coverage', 'mkdocs-material', 'mkdocstrings[python]'],
    #     'export': ['coremltools>=6.0', 'onnx', 'onnxsim', 'onnxruntime', 'openvino-dev>=2022.3'],
    #     'tf': ['onnx2tf', 'sng4onnx', 'tflite_support', 'tensorflow']},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        # 'Intended Audience :: Education',
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        # 'Programming Language :: Python :: 3.11',
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Operating System :: POSIX :: Linux",
        # 'Operating System :: MacOS',
        # 'Operating System :: Microsoft :: Windows',
    ],
    keywords="machine-learning, deep-learning, vision, ML, DL, AI, YOLO, Ultralytics, SNUAILAB",
    entry_points={
        "console_scripts": [
            "waffle_dataset = waffle_dough.dataset.cli:main",
            "waffle_dough = waffle_dough.dataset.cli:main",
            "wd = waffle_dough.dataset.cli:main",
        ]
    },
)



import logging
import traceback


class BaseException(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        stack = traceback.extract_stack()
        logger = logging.getLogger(stack[-2].filename)
        logger.error(self.__str__(), stacklevel=2)



from itertools import groupby
from typing import Any, Union

import cv2
import numpy as np
from pycocotools import mask as mask_utils
from shapely.geometry import Polygon as ShapelyPolygon

from waffle_dough.math.box import convert_box
from waffle_dough.type.annotation_type import BoxType, SegmentationType


def mask_to_polygon(mask: np.ndarray) -> list[list]:
    """Convert mask to polygon.
    Args:
        mask (np.ndarray): mask image.

    Returns:
        list[list]: polygon coordinates. [[x1, y1, x2, y2, ...], ...]
    """
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    polygons = []

    for contour in contours:
        polygon = []
        contour = contour.reshape(-1, 2)
        for point in contour:
            x, y = point
            if x != 0:
                x += 1
            if y != 0:
                y += 1
            polygon.extend([x, y])
        polygons.append(polygon)
    return polygons


def mask_to_rle(mask: np.ndarray) -> dict:
    """Convert mask to rle.
    Args:
        mask (np.ndarray): mask image.

    Returns:
        dict: rle.
    """
    rle = {"counts": [], "size": list(mask.shape)}
    counts = rle.get("counts")
    for i, (value, elements) in enumerate(groupby(mask.ravel(order="F"))):
        if i == 0 and value != 0:
            counts.append(0)
        counts.append(len(list(elements)))
    return rle


def polygon_to_mask(
    polygons: list[list[float]], image_size: tuple, fill_value: int = 255
) -> np.ndarray:
    """Convert polygon to mask.
    Args:
        polygons (list[float]): polygons coordinates. [[x1, y1, x2, y2, x3, y3, ...], []]
        image_size (tuple): image size. (height, width)
        fill_value (int): fill value for mask.

    Returns:
        np.ndarray: mask image.
    """
    mask = np.zeros(image_size, dtype=np.uint8)
    for polygon in polygons:
        points = np.array(polygon).reshape(-1, 2).astype(np.int32)
        cv2.fillPoly(mask, [points], fill_value)
    return mask


def polygon_to_rle(polygons: list[float], image_size: tuple) -> dict:
    """Convert polygon to rle.
    Args:
        polygons (list[float]): polygons coordinates. [[x1, y1, x2, y2, x3, y3, ...], []]
        image_size (tuple): image size. (height, width)

    Returns:
        dict: rle.
    """
    mask = polygon_to_mask(polygons, image_size)
    return mask_to_rle(mask)


def rle_to_mask(rle: dict, fill_value: int = 255) -> np.ndarray:
    """Convert rle to mask.
    Args:
        rle (dict): rle.
        fill_value (int): fill value for mask.

    Returns:
        np.ndarray: mask image.
    """
    mask = mask_utils.frPyObjects(rle, rle["size"][0], rle["size"][1])  # height, width
    return mask_utils.decode(mask) * fill_value


def rle_to_polygon(rle: dict) -> list[float]:
    """Convert rle to polygon.
    Args:
        rle (dict): rle.

    Returns:
        list[float]: polygon coordinates. [x1, y1, x2, y2, x3, y3, ...]
    """
    mask = rle_to_mask(rle)
    return mask_to_polygon(mask)


def convert_segmentation(
    segmentation: Any,
    src_type: Union[str, SegmentationType],
    dst_type: Union[str, SegmentationType],
    image_size: tuple[int, int] = None,
) -> Union[list[list[float]], dict, np.ndarray]:
    """Convert segmentation from src_type to dst_type.
    Args:
        segmentation (list[list[float]]): segmentation
        src_type (Union[str, SegmentationType]): source type
        dst_type (Union[str, SegmentationType]): destination type
        image_size (Optional(tuple[int, int])): image size for polygon type. (width, height)

    Returns:
        Union[list[list[float]], dict, np.ndarray]: converted segmentation
    """

    if src_type == dst_type:
        return segmentation

    if src_type == SegmentationType.POLYGON:
        if image_size is None:
            raise ValueError(
                "image_size must be provided when converting from polygon to other types."
            )
        if dst_type == SegmentationType.MASK:
            return polygon_to_mask(segmentation, image_size)
        elif dst_type == SegmentationType.RLE:
            return polygon_to_rle(segmentation, image_size)
    elif src_type == SegmentationType.MASK:
        if dst_type == SegmentationType.POLYGON:
            return mask_to_polygon(segmentation)
        elif dst_type == SegmentationType.RLE:
            return mask_to_rle(segmentation)
    elif src_type == SegmentationType.RLE:
        if dst_type == SegmentationType.POLYGON:
            return rle_to_polygon(segmentation)
        elif dst_type == SegmentationType.MASK:
            return rle_to_mask(segmentation)
    else:
        raise NotImplementedError(f"Cannot convert segmentation from {src_type} to {dst_type}")


def get_segmentation_area(
    segmentation: list[list[float]],
    segmentation_type: Union[str, SegmentationType],
    image_size: tuple[int, int] = None,
) -> float:
    """Get segmentation area.
    Args:
        segmentation (list[list[float]]): segmentation
        segmentation_type (Union[str, SegmentationType]): segmentation type
        image_size (Optional(tuple[int, int])): image size for polygon type. (width, height)

    Returns:
        float: segmentation area
    """
    polygons = convert_segmentation(
        segmentation, segmentation_type, SegmentationType.POLYGON, image_size=image_size
    )
    area = 0
    for polygon in polygons:
        polygon = ShapelyPolygon(np.array(polygon).reshape(-1, 2))
        area += polygon.area
    return area


def get_segmentation_box(
    segmentation: list[list[float]],
    segmentation_type: Union[str, SegmentationType],
    image_size: tuple[int, int] = None,
    box_type: Union[str, BoxType] = BoxType.XYWH,
) -> list[float]:
    """Get segmentation box.
    Args:
        segmentation (list[list[float]]): segmentation
        segmentation_type (Union[str, SegmentationType]): segmentation type
        image_size (Optional(tuple[int, int])): image size for polygon type. (width, height)
        box_type (Union[str, BoxType]): box type. Default to BoxType.XYWH.

    Returns:
        list[float]: segmentation bbox [x1, y1, x2, y2]
    """
    polygons = convert_segmentation(
        segmentation, segmentation_type, SegmentationType.POLYGON, image_size=image_size
    )
    x1, y1, x2, y2 = np.inf, np.inf, -np.inf, -np.inf
    for polygon in polygons:
        polygon = np.array(polygon).reshape(-1, 2)
        x1 = min(x1, polygon[:, 0].min())
        y1 = min(y1, polygon[:, 1].min())
        x2 = max(x2, polygon[:, 0].max())
        y2 = max(y2, polygon[:, 1].max())
    return convert_box([x1, y1, x2, y2], BoxType.XYXY, box_type)



from sqlalchemy import JSON, Boolean, Column, Double, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from .base import Base


class Annotation(Base):
    __tablename__ = "annotations"

    id = Column(String, primary_key=True)
    task = Column(String)
    image_id = Column(String, ForeignKey("images.id"))
    category_id = Column(String, ForeignKey("categories.id"))
    bbox = Column(JSON)
    segmentation = Column(JSON)
    area = Column(Double)
    keypoints = Column(JSON)
    num_keypoints = Column(Integer)
    caption = Column(String)
    value = Column(Double)
    iscrowd = Column(Integer)
    score = Column(Double)

    image = relationship("Image", back_populates="annotations")
    category = relationship("Category", back_populates="annotations")



from typing import ClassVar, Optional, Union

from pydantic import BaseModel, Field, field_validator

from waffle_dough.field.base_field import BaseField
from waffle_dough.field.validator.annotation_validator import (
    validate_area,
    validate_bbox,
    validate_caption,
    validate_iscrowd,
    validate_keypoints,
    validate_num_keypoints,
    validate_score,
    validate_segmentation,
    validate_value,
)
from waffle_dough.math.box import get_box_area
from waffle_dough.math.segmentation import (
    get_segmentation_area,
    get_segmentation_box,
)
from waffle_dough.type import BoxType, SegmentationType, TaskType


class AnnotationInfo(BaseField):
    image_id: str = Field(...)
    category_id: str = Field(None)
    bbox: Optional[list[float]] = Field(None)
    segmentation: Optional[Union[dict, list[list[float]]]] = Field(None)
    area: Optional[float] = Field(None)
    keypoints: Optional[list[float]] = Field(None)
    num_keypoints: Optional[int] = Field(None)
    caption: Optional[str] = Field(None)
    value: Optional[float] = Field(None)
    iscrowd: Optional[int] = Field(None)
    score: Optional[float] = Field(None)

    extra_required_fields: ClassVar[dict[TaskType, list[str]]] = {
        TaskType.CLASSIFICATION: ["image_id", "category_id"],
        TaskType.OBJECT_DETECTION: ["image_id", "category_id", "bbox"],
        TaskType.SEMANTIC_SEGMENTATION: ["image_id", "category_id", "segmentation"],
        TaskType.INSTANCE_SEGMENTATION: ["image_id", "category_id", "segmentation"],
        TaskType.KEYPOINT_DETECTION: ["image_id", "category_id", "keypoints"],
        TaskType.TEXT_RECOGNITION: ["image_id", "caption"],
        TaskType.REGRESSION: ["image_id", "category_id", "value"],
    }

    @field_validator("bbox")
    def check_bbox(cls, v):
        return validate_bbox(v)

    @field_validator("segmentation")
    def check_segmentation(cls, v):
        return validate_segmentation(v)

    @field_validator("area")
    def check_area(cls, v):
        return validate_area(v)

    @field_validator("keypoints")
    def check_keypoints(cls, v):
        return validate_keypoints(v)

    @field_validator("num_keypoints")
    def check_num_keypoints(cls, v):
        return validate_num_keypoints(v)

    @field_validator("value")
    def check_value(cls, v):
        return validate_value(v)

    @field_validator("caption")
    def check_caption(cls, v):
        return validate_caption(v)

    @field_validator("iscrowd")
    def check_iscrowd(cls, v):
        return validate_iscrowd(v)

    @field_validator("score")
    def check_score(cls, v):
        return validate_score(v)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_default_values()

    def set_default_values(self):
        # init default values
        if self.bbox is None:
            if self.segmentation is not None:
                self.bbox = get_segmentation_box(self.segmentation, SegmentationType.POLYGON)

        if self.area is None:
            if self.segmentation is not None:
                self.area = get_segmentation_area(self.segmentation, SegmentationType.POLYGON)
            elif self.bbox is not None:
                self.area = get_box_area(self.bbox, BoxType.XYWH)

        if self.iscrowd is None and self.bbox is not None:
            self.iscrowd = 0

        if self.num_keypoints is None and self.keypoints is not None:
            self.num_keypoints = len(self.keypoints) // 3

    @classmethod
    def classification(
        cls,
        image_id: str,
        category_id: str,
        score: float = None,
    ) -> "AnnotationInfo":
        """Classification Annotation Format

        Args:
            image_id (str): image id.
            category_id (str): category id.
            score (float, optional): prediction score. Default to None.

        Returns:
            Annotation: annotation class
        """
        return cls(
            task=TaskType.CLASSIFICATION,
            image_id=image_id,
            category_id=category_id,
            score=score,
        )

    @classmethod
    def object_detection(
        cls,
        image_id: str,
        category_id: str,
        bbox: list[float],
        area: int = None,
        iscrowd: int = 0,
        score: float = None,
    ) -> "AnnotationInfo":
        """Object Detection Annotation Format

        Args:
            image_id (str): image id.
            category_id (str): category id.
            bbox (list[float]): [x1, y1, w, h].
            area (int): bbox area.
            iscrowd (int, optional): is crowd or not. Default to 0.
            score (float, optional): prediction score. Default to None.

        Returns:
            Annotation: annotation class
        """
        return cls(
            task=TaskType.OBJECT_DETECTION,
            image_id=image_id,
            category_id=category_id,
            bbox=bbox,
            area=area,
            iscrowd=iscrowd,
            score=score,
        )

    @classmethod
    def semantic_segmentation(
        cls,
        image_id: str,
        category_id: str,
        segmentation: Union[list[list[float]], dict],
        bbox: list[float] = None,
        area: int = None,
        iscrowd: int = 0,
        score: float = None,
    ) -> "AnnotationInfo":
        """Segmentation Annotation Format

        Args:
            image_id (str): image id.
            category_id (str): category id.
            bbox (list[float]): [x1, y1, w, h].
            segmentation (Union[list[list[float]], dict]): [[x1, y1, x2, y2, x3, y3, ...], [polygon]] or RLE.
            area (int): segmentation segmentation area.
            iscrowd (int, optional): is crowd or not. Default to 0.
            score (float, optional): prediction score. Default to None.

        Returns:
            Annotation: annotation class
        """
        return cls(
            task=TaskType.SEMANTIC_SEGMENTATION,
            image_id=image_id,
            category_id=category_id,
            bbox=bbox,
            segmentation=segmentation,
            area=area,
            iscrowd=iscrowd,
            score=score,
        )

    @classmethod
    def instance_segmentation(
        cls,
        image_id: str,
        category_id: str,
        segmentation: Union[list[list[float]], dict],
        bbox: list[float] = None,
        area: int = None,
        iscrowd: int = 0,
        score: float = None,
    ) -> "AnnotationInfo":
        """Instance Annotation Format

        Args:
            image_id (str): image id.
            category_id (str): category id.
            bbox (list[float]): [x1, y1, w, h].
            segmentation (Union[list[list[float]], dict]): [[x1, y1, x2, y2, x3, y3, ...], [polygon]] or RLE.
            area (int): segmentation segmentation area.
            iscrowd (int, optional): is crowd or not. Default to 0.
            score (float, optional): prediction score. Default to None.

        Returns:
            Annotation: annotation class
        """
        return cls(
            task=TaskType.INSTANCE_SEGMENTATION,
            image_id=image_id,
            category_id=category_id,
            bbox=bbox,
            segmentation=segmentation,
            area=area,
            iscrowd=iscrowd,
            score=score,
        )

    @classmethod
    def keypoint_detection(
        cls,
        image_id: str,
        category_id: str,
        keypoints: list[float],
        bbox: list[float],
        num_keypoints: int = None,
        area: int = None,
        segmentation: list[list[float]] = None,
        iscrowd: int = 0,
        score: list[float] = None,
    ) -> "AnnotationInfo":
        """Keypoint Detection Annotation Format

        Args:
            image_id (str): image id.
            category_id (str): category id.
            bbox (list[float]): [x1, y1, w, h].
            keypoints (list[float]):
                [x1, y1, v1(visible flag), x2, y2, v2(visible flag), ...].
                visible flag is one of [0(Not labeled), 1(Labeled but not visible), 2(labeled and visible)]
            num_keypoints: number of labeled keypoints
            area (int): segmentation segmentation or bbox area.
            segmentation (list[list[float]], optional): [[x1, y1, x2, y2, x3, y3, ...], [polygon]].
            iscrowd (int, optional): is crowd or not. Default to 0.
            score (list[float], optional): prediction scores. Default to None.

        Returns:
            Annotation: annotation class
        """
        return cls(
            task=TaskType.KEYPOINT_DETECTION,
            image_id=image_id,
            category_id=category_id,
            bbox=bbox,
            keypoints=keypoints,
            num_keypoints=num_keypoints,
            segmentation=segmentation,
            area=area,
            iscrowd=iscrowd,
            score=score,
        )

    @classmethod
    def regression(cls, image_id: str, category_id: str, value: float) -> "AnnotationInfo":
        """Regression Annotation Format

        Args:
            image_id (str): image id.
            category_id (str): category id.
            value (float): regression value.

        Returns:
            Annotation: annotation class
        """
        return cls(
            task=TaskType.REGRESSION,
            image_id=image_id,
            category_id=category_id,
            value=value,
        )

    @classmethod
    def text_recognition(
        cls,
        image_id: str,
        category_id: str,
        caption: str,
        score: float = None,
    ) -> "AnnotationInfo":
        """Text Recognition Annotation Format

        Args:
            image_id (str): image id.
            category_id (str): category id.
            caption (str): string.
            score (float, optional): prediction score. Default to None.

        Returns:
            Annotation: annotation class
        """
        return cls(
            task=TaskType.TEXT_RECOGNITION,
            image_id=image_id,
            category_id=category_id,
            caption=caption,
            score=score,
        )


class UpdateAnnotationInfo(BaseModel):
    category_id: Optional[str] = Field(None)
    bbox: Optional[list[float]] = Field(None)
    segmentation: Optional[Union[dict, list[list[float]]]] = Field(None)
    area: Optional[float] = Field(None)
    keypoints: Optional[list[float]] = Field(None)
    num_keypoints: Optional[int] = Field(None)
    caption: Optional[str] = Field(None)
    value: Optional[float] = Field(None)
    iscrowd: Optional[int] = Field(None)
    score: Optional[float] = Field(None)

    @field_validator("bbox")
    def check_bbox(cls, v):
        return validate_bbox(v)

    @field_validator("segmentation")
    def check_segmentation(cls, v):
        return validate_segmentation(v)

    @field_validator("area")
    def check_area(cls, v):
        return validate_area(v)

    @field_validator("keypoints")
    def check_keypoints(cls, v):
        return validate_keypoints(v)

    @field_validator("num_keypoints")
    def check_num_keypoints(cls, v):
        return validate_num_keypoints(v)

    @field_validator("value")
    def check_value(cls, v):
        return validate_value(v)

    @field_validator("caption")
    def check_caption(cls, v):
        return validate_caption(v)

    @field_validator("iscrowd")
    def check_iscrowd(cls, v):
        return validate_iscrowd(v)

    @field_validator("score")
    def check_score(cls, v):
        return validate_score(v)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_default_values()

    def set_default_values(self):
        # init default values
        if self.bbox is None:
            if self.segmentation is not None:
                self.bbox = get_segmentation_box(self.segmentation, SegmentationType.POLYGON)

        if self.area is None:
            if self.segmentation is not None:
                self.area = get_segmentation_area(self.segmentation, SegmentationType.POLYGON)
            elif self.bbox is not None:
                self.area = get_box_area(self.bbox, BoxType.XYWH)

        if self.iscrowd is None and self.bbox is not None:
            self.iscrowd = 0

        if self.num_keypoints is None and self.keypoints is not None:
            self.num_keypoints = len(self.keypoints) // 3

    @classmethod
    def classification(
        cls,
        category_id: Optional[str] = None,
        score: Optional[float] = None,
    ) -> "UpdateAnnotationInfo":
        return cls(
            category_id=category_id,
            score=score,
        )

    @classmethod
    def object_detection(
        cls,
        category_id: Optional[str] = None,
        bbox: Optional[list[float]] = None,
        area: Optional[int] = None,
        iscrowd: Optional[int] = None,
        score: Optional[float] = None,
    ) -> "UpdateAnnotationInfo":
        return cls(
            category_id=category_id,
            bbox=bbox,
            area=area,
            iscrowd=iscrowd,
            score=score,
        )

    @classmethod
    def semantic_segmentation(
        cls,
        category_id: Optional[str] = None,
        segmentation: Optional[Union[list[list[float]], dict]] = None,
        bbox: Optional[list[float]] = None,
        area: Optional[int] = None,
        iscrowd: Optional[int] = None,
        score: Optional[float] = None,
    ) -> "UpdateAnnotationInfo":
        return cls(
            category_id=category_id,
            segmentation=segmentation,
            bbox=bbox,
            area=area,
            iscrowd=iscrowd,
            score=score,
        )

    @classmethod
    def instance_segmentation(
        cls,
        category_id: Optional[str] = None,
        segmentation: Optional[Union[list[list[float]], dict]] = None,
        bbox: Optional[list[float]] = None,
        area: Optional[int] = None,
        iscrowd: Optional[int] = None,
        score: Optional[float] = None,
    ) -> "UpdateAnnotationInfo":
        return cls(
            category_id=category_id,
            segmentation=segmentation,
            bbox=bbox,
            area=area,
            iscrowd=iscrowd,
            score=score,
        )

    @classmethod
    def keypoint_detection(
        cls,
        category_id: Optional[str] = None,
        keypoints: Optional[list[float]] = None,
        bbox: Optional[list[float]] = None,
        num_keypoints: Optional[int] = None,
        area: Optional[int] = None,
        segmentation: Optional[list[list[float]]] = None,
        iscrowd: Optional[int] = None,
        score: Optional[list[float]] = None,
    ) -> "UpdateAnnotationInfo":
        return cls(
            category_id=category_id,
            keypoints=keypoints,
            bbox=bbox,
            num_keypoints=num_keypoints,
            area=area,
            segmentation=segmentation,
            iscrowd=iscrowd,
            score=score,
        )

    @classmethod
    def regression(
        cls,
        category_id: Optional[str] = None,
        value: Optional[float] = None,
    ) -> "UpdateAnnotationInfo":
        return cls(
            category_id=category_id,
            value=value,
        )

    @classmethod
    def text_recognition(
        cls,
        category_id: Optional[str] = None,
        caption: Optional[str] = None,
        score: Optional[float] = None,
    ) -> "UpdateAnnotationInfo":
        return cls(
            category_id=category_id,
            caption=caption,
            score=score,
        )
